# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Any
from uuid import UUID

import numpy as np

from ...data import DataType
from ...data.primitive_type_enum import DataTypeEnum
from ..utils import str2uuid, to_tuple
from .property_group import ConcatenatedPropertyGroup


if TYPE_CHECKING:
    from .concatenator import Concatenator
    from .data import ConcatenatedData
    from .drillhole import ConcatenatedDrillhole


class DrillholesGroupTable(ABC):
    """
    A class to store the information of a PropertyGroup.

    :param parent: The parent DrillholesConcatenator.
    :param name: The name of the PropertyGroup to extract.
    """

    def __init__(self, parent: Concatenator, name: str):
        self._association: tuple | None = None
        self._index_by_drillhole: dict | None = None
        self._properties: tuple | None = None
        self._properties_types: tuple | None = None

        self._property_groups: dict[UUID, ConcatenatedPropertyGroup] = (
            self._get_property_groups(parent, name)
        )
        self._parent: Concatenator = parent
        self._name: str = name

    def _create_empty_structured_array(
        self,
        names: tuple[str],
        drillhole: bool = True,
    ) -> np.ndarray:
        """
        Create an empty structured array that can contains the data.

        :param names: The names to extract.
        :param drillhole: If True, the drillholes are added to the table.

        :return: an empty structured array.
        """
        dtypes = [("Drillhole", "O")] if drillhole else []
        no_data_values = [None] if drillhole else []

        properties = self.association + self.properties
        types = (np.float32,) * len(self.association) + self.properties_type

        for name, dtype in zip(properties, types, strict=False):
            if name in names:
                if dtype not in [np.float32, np.int32, np.uint32, bool]:
                    dtype = "O"
                dtypes.append((name, dtype))
                no_data_values.append(self.nan_value_from_name(name))

        empty_array = np.recarray(
            (self.parent.data[self.association[0]].shape[0],), dtype=dtypes
        )

        for name, ndv in zip(empty_array.dtype.names, no_data_values, strict=False):
            empty_array[name].fill(ndv)

        return empty_array

    def _depth_table_by_key(
        self,
        names: tuple,
        drillholes: bool = True,
    ) -> np.ndarray:
        """
        Get a table with all the data associated with depth for every drillhole object.

        The Drillhole name is added at the beginning of the table for every row.
        The table is based on the association and contains nan values if no data is found.

        :param names: The names to extract.
        :param drillholes: If True, the drillholes are added to the table.

        :return: a structured array with all the data.
        """
        if self.index_by_drillhole is None:
            raise ValueError("No drillhole found in the concatenator.")

        output_array = self._create_empty_structured_array(
            names=names, drillhole=drillholes
        )
        current_index = 0
        for object_, data_dict in self.index_by_drillhole.items():
            for name, info in data_dict.items():
                if name in names:
                    temp_array = self.parent.data[name][info[0] : info[0] + info[1]]

                    output_array[name][
                        current_index : current_index + temp_array.shape[0]
                    ] = temp_array

            if drillholes:
                output_array["Drillhole"][
                    current_index : current_index + data_dict[self.association[0]][1]
                ] = object_

            current_index += data_dict[self.association[0]][1]

        # transform to a structured array
        return output_array

    @staticmethod
    def _get_property_groups(parent, name) -> dict[UUID, ConcatenatedPropertyGroup]:
        """
        Get all the property groups with a given name in the concatenator.

        :param parent: the concatenator object.
        :param name: the name to get the property groups for.

        :return: a  dictionary of property groups with drillhole object as Key.
        """
        if not hasattr(parent, "property_group_ids"):
            raise TypeError("The parent must be a Concatenator instance.")

        property_groups: dict[UUID, ConcatenatedPropertyGroup] = {}
        for property_group_uid in parent.property_group_ids:
            property_group = parent.workspace.get_entity(str2uuid(property_group_uid))[
                0
            ]
            if (
                isinstance(property_group, ConcatenatedPropertyGroup)
                and property_group.name == name
            ):
                property_groups[property_group.parent.uid] = property_group

        if not property_groups:
            raise ValueError(
                f"No property group with name '{name}' found in the concatenator."
            )

        return property_groups

    def _get_properties_names_types(self):
        if not self._properties:
            properties_name: tuple = ()
            properties_type: tuple = ()

            for property_group in self.property_groups.values():
                if property_group.properties is None:
                    continue

                for property_ in property_group.properties:
                    property_temp = property_group.parent.get_data(property_)[0]

                    if property_temp.name not in self.association + properties_name:
                        properties_name += (property_temp.name,)
                        properties_type += (
                            DataTypeEnum.from_primitive_type(
                                property_temp.entity_type.primitive_type
                            ),
                        )

            # sort the properties names and dtypes
            mapper = dict(zip(properties_name, properties_type, strict=False))
            properties_name = tuple(sorted(properties_name))
            properties_type = tuple(mapper[name] for name in properties_name)

            if properties_name:
                self._properties = properties_name
                self._properties_types = properties_type

    def add_values_to_property_group(
        self, name: str, values: np.ndarray, data_type: DataType | None = None
    ):
        """
        Push the values to each drillhole of the property group based on association.

        :param name: The name of the data to push.
        :param values: The values to push.
        :param data_type: The data type associated to description;
            useful especially for referenced data.
        """
        if not isinstance(name, str) or name in self.parent.data:
            raise KeyError("The name must be a string not present in data.")

        # ensure the length of the values is the same as the length of the template
        if values.shape != self.parent.data[self.association[0]].shape:
            raise ValueError(
                "The length of the values must be the same as the association "
                f"({self.parent.data[self.association[0]].shape})."
            )

        if not isinstance(data_type, DataType):
            primitive_type = DataType.primitive_type_from_values(values)
            data_type = DataType.find_or_create_type(
                self.parent.workspace, primitive_type, name=name
            )

        for drillhole_uid, indices in self.index_by_drillhole.items():
            # get the drillhole
            drillhole: ConcatenatedDrillhole = self.parent.workspace.get_entity(  # type: ignore
                str2uuid(drillhole_uid)
            )[0]

            # add data to the drillhole
            drillhole.add_data(
                {
                    name: {
                        **{
                            "values": values[
                                indices[self.association[0]][0] : indices[
                                    self.association[0]
                                ][0]
                                + indices[self.association[0]][1]
                            ]
                        },
                        "entity_type": data_type,
                    },
                },
                property_group=self.property_groups[drillhole.uid],
            )

        self._update_drillholes_group_table(name, data_type)

    @property
    def association(self) -> tuple:
        """
        The depth association of the PropertyGroup.
        """
        if self._association is None:
            if self.property_group_type == "Interval table":
                self._association = (
                    list(self.property_groups.values())[0].from_.name,
                    list(self.property_groups.values())[0].to_.name,
                )
            elif self.property_group_type == "Depth table":
                self._association = (
                    list(self.property_groups.values())[0].depth_.name,
                )
            else:
                raise TypeError(
                    f"The property group type '{self.property_group_type}' is not supported."
                )

        return self._association

    @property
    def depth_table(
        self,
    ) -> np.ndarray:
        """
        Get a table with all the data associated with depth for every drillhole object.

        The Drillhole name is added at the beginning of the table for every row.
        The table is based on the association and contains nan values if no data is found.

        :return: a structured array with all the data.
        """
        # get the depth table
        return self._depth_table_by_key(self.association + self.properties, True)

    def depth_table_by_name(
        self,
        names: tuple[str] | str,
        spatial_index: bool = False,
    ) -> np.ndarray:
        """
        Get a table with specific data associated with depth for every drillhole object.

        :param names: The names to extract.
        :param spatial_index: If True, the spatial index is added to the table.

        :return: a table containing the Drillholes, the association and the data.
        """
        names = to_tuple(names)

        # ensure names are in properties
        if not all(name in self.properties for name in names):
            raise KeyError("The names are not in the list of properties.")

        if spatial_index:
            return self._depth_table_by_key(self.association + names, True)

        return self._depth_table_by_key(names, False)

    @property
    def index_by_drillhole(
        self,
    ) -> dict[bytes, dict[str, list[int]]]:
        """
        Get for every object index and count of all the data in 'association' and 'properties'

        :return: A dictionary with the object uuid and the index of all the data.
        """
        if self._index_by_drillhole is None:
            index_by_drillhole: dict[bytes, dict[str, list[int]]] = {}
            names = self.association + self.properties
            for drillhole in np.sort(self._parent.index[names[0]], order="Start index")[
                "Object ID"
            ]:
                index_by_drillhole[drillhole] = {}
                for name in names:
                    if drillhole in self._parent.index[name]["Object ID"]:
                        index_by_drillhole[drillhole][name] = list(
                            self._parent.index[name][
                                self._parent.index[name]["Object ID"] == drillhole
                            ][0]
                        )[:2]
                    else:
                        index_by_drillhole[drillhole][name] = [0, 0]

            if index_by_drillhole:
                self._index_by_drillhole = index_by_drillhole
            else:
                raise AssertionError("No drillhole found in the concatenator.")

        return self._index_by_drillhole

    @property
    def name(self) -> str:
        """
        The name of the PropertyGroup.
        """
        return self._name

    def nan_value_from_name(self, name: str) -> Any:
        """
        Get the nan value of a data from the name.

        :param name: The name of the data to get.

        :return: The nan value.
        """
        if name not in self.properties + self.association:
            raise KeyError(f"The name '{name}' is not in the list of properties.")

        # get the data of the know association
        data: ConcatenatedData = self.parent.workspace.get_entity(  # type: ignore
            str2uuid(self.parent.index[name][0][-2])
        )[0].get_data(str2uuid(self.parent.index[name][0][-1]))[0]

        return data.nan_value

    @property
    def parent(self) -> Concatenator:
        """
        The parent Concatenator object.
        """
        return self._parent

    @property
    def properties(self) -> tuple:
        """
        The names of the associated data.
        """
        if not self._properties:
            self._get_properties_names_types()

        if self._properties is None:
            return ()

        return self._properties

    @property
    def properties_type(self) -> tuple:
        """
        The dtype of the associated data.

        They are ordered in the same way as the properties.
        """
        if not self._properties_types:
            self._get_properties_names_types()

        if self._properties_types is None:
            return ()

        return self._properties_types

    @property
    def property_group_type(self) -> str:
        """
        The type of the PropertyGroup.
        """
        return list(self.property_groups.values())[0].property_group_type

    @property
    def property_groups(self) -> dict[UUID, ConcatenatedPropertyGroup]:
        """
        Get all the property groups in the concatenator.

        :return: A list containing all the property groups.
        """
        return self._property_groups

    def _update_drillholes_group_table(self, name: str, data_type: DataType):
        """
        Update the drillholes group table with a new property group.

        :param name: The name of the property group to update.
        :param data_type: The data type of the property group.
        """
        self.parent.update_data_index()
        self.parent.workspace.update_attribute(self.parent, "concatenated_attributes")
        self._property_groups = self._get_property_groups(self.parent, self.name)

        dtype = DataTypeEnum.from_primitive_type(data_type.primitive_type)

        if self._properties is not None and self._properties_types is not None:
            self._properties += (name,)
            self._properties_types += (dtype,)
        else:
            self._properties = ((name,),)
            self._properties_types = ((dtype,),)

        if self._index_by_drillhole is not None:
            for value in self._index_by_drillhole.values():
                value[name] = value[self.association[0]]



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=too-many-lines

from __future__ import annotations

import uuid
import warnings

import numpy as np
from h5py import special_dtype

from ...data import Data, DataAssociationEnum, DataType
from ...groups import Group
from ..entity import Entity
from ..entity_type import EntityType
from ..utils import INV_KEY_MAP, KEY_MAP, as_str_if_utf8_bytes, as_str_if_uuid, str2uuid
from .concatenated import Concatenated
from .data import ConcatenatedData
from .drillholes_group_table import DrillholesGroupTable
from .object import ConcatenatedObject
from .property_group import ConcatenatedPropertyGroup, PropertyGroup


PROPERTY_KWARGS = {
    "trace": {"maxshape": (None,)},
    "trace_depth": {"maxshape": (None,)},
    "property_group_ids": {
        "dtype": special_dtype(vlen=str),
        "maxshape": (None,),
    },
    "surveys": {"maxshape": (None,)},
}


class Concatenator(Group):  # pylint: disable=too-many-public-methods
    """
    Class modifier for concatenation of objects and data.
    """

    _concat_attr_str: str | None = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._attribute_map.update(
            {
                self.concat_attr_str: "concatenated_attributes",
                "Property Groups IDs": "property_group_ids",
                "Concatenated object IDs": "concatenated_object_ids",
            }
        )

        self._concatenated_attributes: dict | None = None
        self._attributes_keys: list[uuid.UUID] | None = None
        self._concatenated_object_ids: list[bytes] | None = None
        self._property_group_ids: np.ndarray | None = None

        self._data: dict
        self._index: dict

    @property
    def attributes_keys(self) -> list | None:
        """List of uuids present in the concatenated attributes."""
        if getattr(self, "_attributes_keys", None) is None:
            attributes_keys = []
            if self.concatenated_attributes is not None:
                attributes_keys = [
                    elem["ID"] for elem in self.concatenated_attributes["Attributes"]
                ]

            self._attributes_keys = attributes_keys

        return self._attributes_keys

    def add_children(
        self, children: Entity | PropertyGroup | list[Entity | PropertyGroup]
    ) -> None:
        """
        :param children: Add a list of entities as
            :obj:`~geoh5py.shared.entity.Entity.children`
        """
        if not isinstance(children, list):
            children = [children]

        for child in children:
            if not (
                isinstance(child, Concatenated)
                or (
                    isinstance(child, Data)
                    and child.association
                    in (DataAssociationEnum.OBJECT, DataAssociationEnum.GROUP)
                )
            ):
                warnings.warn(
                    f"Expected a Concatenated object, not {type(child).__name__}"
                )
                continue

            if child in self._children:
                continue

            self._children.append(child)

            if (
                not isinstance(child, PropertyGroup)
                and hasattr(child, "parent")
                and child.parent != self
            ):
                child.parent = self

    def add_save_concatenated(self, child) -> None:
        """
        Add or save a concatenated entity.

        :param child: Concatenated entity
        """
        self.update_concatenated_attributes(child)

        if hasattr(child, "values"):
            self.update_array_attribute(child, child.name)
        elif hasattr(child, "surveys"):  # Specific to drillholes
            uid = as_str_if_uuid(child.uid).encode()
            concat_object_ids = [uid]
            if self._concatenated_object_ids is not None:
                if uid not in self._concatenated_object_ids:  # type: ignore
                    concat_object_ids = (
                        self._concatenated_object_ids + concat_object_ids  # type: ignore
                    )
                else:
                    concat_object_ids = self._concatenated_object_ids  # type: ignore

            self.concatenated_object_ids = concat_object_ids
            self.update_array_attribute(child, "surveys")
            self.update_array_attribute(child, "trace")

        child.on_file = True

    @property
    def concat_attr_str(self) -> str:
        """String identifier for the concatenated attributes."""
        if self._concat_attr_str is None:
            self._concat_attr_str = "Attributes"
            if self.workspace.version is not None and self.workspace.version > 2.0:
                self._concat_attr_str = "Attributes Jsons"
        return self._concat_attr_str

    @property
    def concatenated_attributes(self) -> dict | None:
        """Dictionary of concatenated objects and data attributes."""
        if self._concatenated_attributes is None:
            concatenated_attributes = self.workspace.fetch_concatenated_attributes(self)

            if concatenated_attributes is None:
                concatenated_attributes = {"Attributes": []}

            self._concatenated_attributes = concatenated_attributes

        return self._concatenated_attributes

    @concatenated_attributes.setter
    def concatenated_attributes(self, concatenated_attributes: dict):
        if not isinstance(concatenated_attributes, (dict, type(None))):
            raise ValueError(
                "Input 'concatenated_attributes' must be a dictionary or None"
            )

        self._concatenated_attributes = concatenated_attributes
        self.workspace.update_attribute(self, "concatenated_attributes")

    @property
    def concatenated_object_ids(self) -> list[bytes] | None:
        """Dictionary of concatenated objects and data concatenated_object_ids."""
        if getattr(self, "_concatenated_object_ids", None) is None:
            concatenated_object_ids = self.workspace.fetch_array_attribute(
                self, "concatenated_object_ids"
            )
            if isinstance(concatenated_object_ids, np.ndarray):
                concatenated_object_ids = concatenated_object_ids.tolist()

            self._concatenated_object_ids = concatenated_object_ids  # type: ignore

        return self._concatenated_object_ids

    @concatenated_object_ids.setter
    def concatenated_object_ids(self, object_ids: list[bytes] | None):
        if isinstance(object_ids, np.ndarray):
            object_ids = object_ids.tolist()

        if not isinstance(object_ids, (list, type(None))):
            raise AttributeError(
                "Input value for 'concatenated_object_ids' must be of type list."
            )

        self._concatenated_object_ids = object_ids
        self.workspace.update_attribute(self, "concatenated_object_ids")

    def copy(
        self,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        **kwargs,
    ):
        """
        Function to copy an entity to a different parent entity.

        :param parent: Target parent to copy the entity under. Copied to current
            :obj:`~geoh5py.shared.entity.Entity.parent` if None.
        :param copy_children: Create copies of all children entities along with it.
        :param mask: Array of indices to sub-sample the input entity.
        :param clear_cache: Clear array attributes after copy.

        :return entity: Registered Entity to the workspace.
        """
        if mask is not None:
            warnings.warn("Masking is not supported for Concatenated objects.")

        new_entity: Concatenator = super().copy(  # mypy: ignore-errors
            parent=parent,
            copy_children=False,
            clear_cache=clear_cache,
            omit_list=[
                "_data",
                "_index",
            ],
            **kwargs,
        )

        if not copy_children or self.concatenated_attributes is None:
            return new_entity

        if (
            mask is None and new_entity.workspace != self.workspace
        ):  # Fast copy to new workspace
            new_entity.concatenated_attributes = self.concatenated_attributes
            new_entity.concatenated_object_ids = self.concatenated_object_ids

            for field in self.index:
                values = self.workspace.fetch_concatenated_values(self, field)
                if isinstance(values, tuple):
                    new_entity.data[field], new_entity.index[field] = values

                new_entity.save_attribute(field)

                # Copy over the data type
            for elem in self.concatenated_attributes["Attributes"]:
                if "Name" in elem and "Type ID" in elem:
                    attr_type = self.workspace.fetch_type(
                        uuid.UUID(elem["Type ID"]), "Data"
                    )
                    primitive_type = attr_type.pop("primitive_type")
                    data_type = DataType.find_or_create_type(
                        new_entity.workspace,
                        primitive_type,
                        **attr_type,
                    )
                    new_entity.workspace.save_entity_type(data_type)

            new_entity.workspace.fetch_children(new_entity)
            for child in self.children:
                if not isinstance(child, Concatenated):
                    child.copy(parent=new_entity)
        else:
            for child in self.children:
                child.copy(
                    parent=new_entity, clear_cache=clear_cache, omit_list=["_uid"]
                )

        return new_entity

    def update_data_index(self):
        """
        Update the concatenated data and index of the concatenator
        """
        self._data, self._index = self.fetch_concatenated_data_index()

    @property
    def data(self) -> dict:
        """
        Concatenated data values stored as a dictionary.
        """
        if getattr(self, "_data", None) is None:
            self.update_data_index()

        return self._data

    @data.setter
    def data(self, data: dict):
        if not isinstance(data, dict):
            raise ValueError("Input 'data' must be a dictionary")

        self._data = data

    def delete_index_data(self, label: str, index: int) -> None:
        start, size = self.index[label][index][0], self.index[label][index][1]
        self.data[label] = np.delete(
            self.data[label], np.arange(start, start + size), axis=0
        )
        # Shift indices
        self.index[label]["Start index"][self.index[label]["Start index"] > start] -= (
            size
        )
        self.index[label] = np.delete(self.index[label], index, axis=0)

    def fetch_concatenated_data_index(self):
        """Extract concatenation arrays."""
        data, index = {}, {}
        data_list = self.workspace.fetch_concatenated_list(self, "Index")

        if data_list is not None:
            for field in data_list:
                name = field.replace("\u2044", "/")
                values = self.workspace.fetch_concatenated_values(self, field)
                if isinstance(values, tuple):
                    data[name], index[name] = values

        return data, index

    def fetch_concatenated_objects(self) -> dict:
        """
        Load all concatenated children.
        """
        attr_dict = {}
        if self.concatenated_object_ids is None:
            return {}

        for key in self.concatenated_object_ids:
            attrs = {
                attr: val
                for attr, val in self.get_concatenated_attributes(key).items()
                if "Property" not in attr
            }
            attrs["parent"] = self
            attr_dict[key] = self.workspace.create_from_concatenation(attrs)

        return attr_dict

    def fetch_index(
        self, entity: ConcatenatedObject | ConcatenatedData | EntityType, field: str
    ) -> int | None:
        """
        Fetch the array index for specific concatenated object and data field.

        :param entity: Parent entity with data
        :param field: Name of the target data.
        """
        field = KEY_MAP.get(field, field)

        if field not in self.index:
            return None

        uid = as_str_if_uuid(entity.uid).encode()

        if isinstance(entity, ConcatenatedData):
            ind = np.where(self.index[field]["Data ID"] == uid)[0]
            if len(ind) == 1:
                return ind[0]
        else:
            ind = np.where(self.index[field]["Object ID"] == uid)[0]
            if len(ind) == 1:
                return ind[0]

        return None

    def fetch_start_index(
        self, entity: ConcatenatedObject | ConcatenatedData, label: str
    ) -> int:
        """
        Fetch starting index for a given entity and label.
        Existing date is removed such that new entries can be appended.

        :param entity: Concatenated entity to be added.
        :param label: Name of the attribute requiring an update.
        """
        index = self.fetch_index(entity, label)
        if index is not None:  # First remove the old data
            self.delete_index_data(label, index)
            start = self.data[label].shape[0]

        elif label in self.index:
            start = np.sum(self.index[label]["Size"])
        else:
            start = 0

        return start

    def fetch_values(
        self, entity: ConcatenatedObject | ConcatenatedData | EntityType, field: str
    ) -> np.ndarray | None:
        """
        Get an array of values from concatenated data.

        :param entity: Parent entity with data
        :param field: Name of the target data.
        """
        field = KEY_MAP.get(field, field)

        index = self.fetch_index(entity, field)

        if index is None:
            return None

        start, size = self.index[field][index][0], self.index[field][index][1]

        return self.data[field][start : start + size]

    def get_concatenated_attributes(self, uid: bytes | str | uuid.UUID) -> dict:
        """
        Fast reference index to concatenated attribute keys.
        """
        if self.concatenated_attributes is None:
            return {}

        uid = as_str_if_utf8_bytes(uid)

        if isinstance(uid, str):
            uid = uuid.UUID(uid)

        uid = as_str_if_utf8_bytes(as_str_if_uuid(uid))

        if self.attributes_keys is not None and uid in self.attributes_keys:
            index = self.attributes_keys.index(uid)
        else:
            if self.attributes_keys is not None:
                self.attributes_keys.append(uid)

            if self.concatenated_attributes is not None:
                self.concatenated_attributes["Attributes"].append({})

            index = -1

        return self.concatenated_attributes["Attributes"][index]

    @property
    def index(self) -> dict:
        """
        Concatenated index stored as a dictionary.
        """
        if getattr(self, "_index", None) is None:
            self.update_data_index()

        return self._index

    @index.setter
    def index(self, index: dict):
        if not isinstance(index, dict):
            raise ValueError("Input 'index' must be a dictionary")

        self._index = index

    @property
    def property_group_ids(self) -> list | None:
        """Dictionary of concatenated objects and data property_group_ids."""
        if not self._property_group_ids:
            property_groups_ids = self.workspace.fetch_concatenated_values(
                self, "property_group_ids"
            )

            if property_groups_ids is not None:
                self._property_group_ids = property_groups_ids[0].tolist()

        return self._property_group_ids

    def remove_children(self, children: list | Concatenated):
        """
        Remove children from object.

        This method calls the ObjectBase parent class to remove children from the
        object children, but also deletes the children from the workspace.

        :param children: List of children to remove.
        """
        if not isinstance(children, list):
            children = [children]

        for child in children:
            if child not in self._children:
                continue

            self.remove_entity(child)

    def remove_entity(
        self, entity: ConcatenatedObject | ConcatenatedData | ConcatenatedPropertyGroup
    ):
        """Remove a concatenated entity."""
        if isinstance(entity, ConcatenatedData):
            parent = entity.parent
            # Remove the rows of data and index
            self.update_array_attribute(entity, entity.name, remove=True)
            # Remove the data from the group

            if entity.property_group is not None:
                entity.property_group.remove_properties([entity])

            # Remove from the concatenated Attributes
            parent_attr = self.get_concatenated_attributes(parent.uid)
            name = entity.name
            del parent_attr[f"Property:{name}"]

        elif isinstance(entity, ConcatenatedObject):
            # First remove the children
            entity.remove_children(entity.children.copy())
            object_ids = self.concatenated_object_ids

            if object_ids is not None:
                object_ids.remove(as_str_if_uuid(entity.uid).encode())
                self.concatenated_object_ids = object_ids

        elif isinstance(entity, ConcatenatedPropertyGroup):
            # Remove all data within the group
            parent = entity.parent
            if entity.properties is not None and len(entity.properties) > 0:
                data = [entity.parent.get_entity(uid)[0] for uid in entity.properties]
                entity.parent.remove_children(data)

            if (
                entity.parent.property_groups is not None
                and entity in entity.parent.property_groups
            ):
                entity.parent.property_groups.remove(entity)
            self.update_array_attribute(parent, "property_groups")

        if (
            self.concatenated_attributes is not None
            and self.attributes_keys is not None
        ):
            attr_handle = self.get_concatenated_attributes(entity.uid)
            self.attributes_keys.remove(as_str_if_uuid(entity.uid))
            self.concatenated_attributes["Attributes"].remove(attr_handle)
            self.workspace.repack = True

    def save_attribute(self, field: str):
        """
        Save a concatenated attribute.

        :param field: Name of the attribute
        """
        field = INV_KEY_MAP.get(field, field)
        alias = KEY_MAP.get(field, field)
        self.workspace.update_attribute(self, "index", alias)

        if field in PROPERTY_KWARGS:  # For group property
            if field == "property_groups":
                field = "property_group_ids"

            self.workspace.update_attribute(
                self,
                field,
                values=self.data.get(alias),
                **PROPERTY_KWARGS.get(field, {}),
            )
        else:  # For data values
            self.workspace.update_attribute(self, "data", alias)

    def update_attributes(
        self, entity: ConcatenatedObject | ConcatenatedData, label: str
    ) -> None:
        """
        Update a concatenated entity.
        """
        if label == "attributes":
            self.update_concatenated_attributes(entity)
        elif label == "property_groups":
            if (
                not isinstance(entity, ConcatenatedObject)
                or entity.property_groups is None
            ):
                return

            for prop_group in entity.property_groups:
                self.add_save_concatenated(prop_group)
                if (
                    self.property_group_ids is not None
                    and as_str_if_uuid(prop_group.uid).encode()
                    not in self.property_group_ids
                ):
                    self.property_group_ids.append(
                        as_str_if_uuid(prop_group.uid).encode()
                    )

            self.update_array_attribute(entity, label)

        else:
            if isinstance(entity, Data):
                label = entity.name

            self.update_array_attribute(entity, label)

    def update_concatenated_attributes(
        self, entity: ConcatenatedObject | ConcatenatedData
    ) -> None:
        """
        Update the concatenated attributes.
        :param entity: Concatenated entity with attributes.
        """
        target_attributes = self.get_concatenated_attributes(entity.uid)

        for key, attr in entity.attribute_map.items():
            val = getattr(entity, attr, None)

            if val is None or attr == "property_groups":
                continue

            if isinstance(val, np.ndarray):
                val = "{" + ", ".join(str(e) for e in val.tolist()) + "}"
            elif isinstance(val, uuid.UUID):
                val = as_str_if_uuid(val)
            elif isinstance(val, list):
                val = [as_str_if_uuid(uid) for uid in val]
            elif attr == "association":
                val = val.name.lower().capitalize()

            target_attributes[key] = val

        if isinstance(entity, Data):
            target_attributes["Type ID"] = as_str_if_uuid(entity.entity_type.uid)
        elif hasattr(entity, "properties"):
            pass
        else:
            target_attributes["Object Type ID"] = as_str_if_uuid(entity.entity_type.uid)
        self.workspace.repack = True

    def update_array_attribute(
        self, entity: ConcatenatedObject | ConcatenatedData, field: str, remove=False
    ) -> None:
        """
        Update values stored as data.

        Row data and indices are first removed then appended.

        :param entity: Concatenated entity with array values.
        :param field: Name of the valued field.
        :param remove: Remove the data from the concatenated array.
        """
        if hasattr(entity, f"_{field}"):
            values = getattr(entity, f"_{field}", None)
            obj_id = as_str_if_uuid(entity.uid).encode()
            data_id = as_str_if_uuid(uuid.UUID(int=0)).encode()
        elif entity.name == field:
            values = getattr(entity, "values", None)
            obj_id = as_str_if_uuid(entity.parent.uid).encode()
            data_id = as_str_if_uuid(entity.uid).encode()
        else:
            raise UserWarning(
                f"Input entity {entity} does not have a property or values "
                f"for the requested field {field}"
            )

        if field == "property_groups" and isinstance(values, list):
            field = "property_group_ids"
            values = [as_str_if_uuid(val.uid).encode() for val in values]

        alias = KEY_MAP.get(field, field)

        start = self.fetch_start_index(entity, alias)

        if values is not None and not remove:
            indices = np.hstack(
                [
                    np.core.records.fromarrays(
                        (start, len(values), obj_id, data_id),
                        dtype=[
                            ("Start index", "<u4"),
                            ("Size", "<u4"),
                            ("Object ID", special_dtype(vlen=str)),
                            ("Data ID", special_dtype(vlen=str)),
                        ],
                    )
                ]
            )
            if alias in self.index:
                indices = np.hstack([self.index[alias], indices]).astype(
                    self.index[alias].dtype
                )

            self.index[alias] = indices

            if alias in self.data:
                values = np.hstack([self.data[alias], values])

            self.data[alias] = values

        self.save_attribute(field)

    @property
    def drillholes_tables(self) -> dict:
        """
        Dictionary of drillholes tables.
        Always recompute the drillholes tables to ensure changes.
        """
        drillholes_tables = {}
        if self.property_group_ids is not None:
            for property_group_uid in self.property_group_ids:
                property_group = self.workspace.get_entity(
                    str2uuid(property_group_uid)
                )[0]

                if (
                    property_group is not None
                    and property_group.name not in drillholes_tables
                    and getattr(property_group, "property_group_type", None)
                    in ["Depth table", "Interval table"]
                ):
                    drillholes_tables[property_group.name] = DrillholesGroupTable(
                        self, property_group.name
                    )

        return drillholes_tables

    @property
    def drillholes_table_from_data_name(self) -> dict | dict[str, DrillholesGroupTable]:
        """
        Dictionary of the drillholes tables from data name as keys
        """
        property_group_from_data_name = {}

        for drillholes_table in self.drillholes_tables.values():
            # prepare the data to pass
            data: tuple = ()
            if drillholes_table.association is not None:
                data += drillholes_table.association
            if drillholes_table.properties is not None:
                data += drillholes_table.properties

            for data_name in data:
                property_group_from_data_name[data_name] = drillholes_table

        return property_group_from_data_name



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=R0904

from __future__ import annotations

import uuid
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from ..shared.utils import map_attributes, str2uuid, str_json_to_dict, to_list
from .entity_type import EntityType


if TYPE_CHECKING:  # pragma: no cover
    from .. import shared
    from ..shared.entity_container import EntityContainer
    from ..workspace import Workspace

DEFAULT_CRS = {"Code": "Unknown", "Name": "Unknown"}


class Entity(ABC):  # pylint: disable=too-many-instance-attributes
    """
    Base entity class for Objects, Groups and Data.

    :param entity_type: Entity type registered by the Workspace.
    :param allow_delete: Entity can be deleted from the workspace.
    :param allow_move: Entity can change :obj:`~geoh5py.shared.entity.Entity.parent`
    :param allow_rename: Entity can change name
    :param clipping_ids: List of clipping uuids
    :param metadata: Metadata attached to the entity.
    :param name: Name of the entity
    :param on_file: Whether this Entity is already stored on
        :obj:`~geoh5py.workspace.workspace.Workspace.h5file`.
    :param partially_hidden: Whether this Entity is partially hidden.
    :param parent: Parent entity.
    :param public: Whether this Entity is accessible in the workspace tree and other parts
        of the user interface in ANALYST.
    :param uid: Unique identifier of the entity.
    :param visible: Whether the Entity is visible in camera (checked in ANALYST object tree).
    """

    _attribute_map: dict = {
        "Allow delete": "allow_delete",
        "Allow move": "allow_move",
        "Allow rename": "allow_rename",
        "Clipping IDs": "clipping_ids: list | None",
        "ID": "uid",
        "Name": "name",
        "Partially hidden": "partially_hidden",
        "Public": "public",
        "Visible": "visible",
    }
    _default_name: str | None = None

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        entity_type: shared.EntityType | None = None,
        allow_delete: bool = True,
        allow_move: bool = True,
        allow_rename: bool = True,
        clipping_ids: list[uuid.UUID] | None = None,
        metadata: dict | None = None,
        name: str | None = None,
        on_file: bool = False,
        partially_hidden: bool = False,
        parent: EntityContainer | None = None,
        public: bool = True,
        uid: uuid.UUID | None = None,
        visible: bool = True,
        **kwargs,
    ):
        self.on_file = on_file
        self.name = name or self._default_name or type(self).__name__
        self._entity_type = self.validate_entity_type(entity_type)
        self.uid: uuid.UUID = uid or uuid.uuid4()
        self.allow_delete = allow_delete
        self.allow_move = allow_move
        self.allow_rename = allow_rename
        self.clipping_ids = clipping_ids
        self.metadata = metadata
        self.parent = parent or self.workspace.root
        self.partially_hidden = partially_hidden
        self.public = public
        self.visible = visible

        # TODO Deprecate in favor of explicit attribute setter
        map_attributes(self, **kwargs)
        self.workspace.register(self)

    @property
    def allow_delete(self) -> bool:
        """
        :obj:`bool` Entity can be deleted from the workspace.
        """
        return self._allow_delete

    @allow_delete.setter
    def allow_delete(self, value: bool):
        self._allow_delete = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def allow_move(self) -> bool:
        """
        :obj:`bool` Entity can change :obj:`~geoh5py.shared.entity.Entity.parent`
        """
        return self._allow_move

    @allow_move.setter
    def allow_move(self, value: bool):
        self._allow_move = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def allow_rename(self) -> bool:
        """
        :obj:`bool` Entity can change name
        """
        return self._allow_rename

    @allow_rename.setter
    def allow_rename(self, value: bool):
        self._allow_rename = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def attribute_map(self) -> dict:
        """
        :obj:`dict` Correspondence map between property names used in geoh5py and
        geoh5.
        """
        return self._attribute_map

    @property
    def clipping_ids(self) -> list[uuid.UUID] | None:
        """
        List of clipping uuids.
        """
        return self._clipping_ids

    @clipping_ids.setter
    def clipping_ids(self, value: list | None):
        msg = (
            "Input clipping_ids must be a list of uuid.UUID or None of Slicer objects. "
            f"Provided value of type '{type(value)}'."
        )

        if value is not None:
            verified_values = []
            for val in to_list(value):
                val = self.workspace.get_entity(str2uuid(val))[0]
                if getattr(val, "_default_name", None) != "Slicer":
                    raise TypeError(msg)
                verified_values.append(val.uid)
            value = verified_values

        self._clipping_ids = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def coordinate_reference_system(self) -> dict:
        """
        Coordinate reference system attached to the entity.
        """
        coordinate_reference_system = DEFAULT_CRS

        if self.metadata is not None and "Coordinate Reference System" in self.metadata:
            coordinate_reference_system = self.metadata[
                "Coordinate Reference System"
            ].get("Current", DEFAULT_CRS)

        return coordinate_reference_system

    @coordinate_reference_system.setter
    def coordinate_reference_system(self, value: dict):
        # assert value is a dictionary containing "Code" and "Name" keys
        if not isinstance(value, dict):
            raise TypeError("Input coordinate reference system must be a dictionary")

        if value.keys() != {"Code", "Name"}:
            raise KeyError(
                "Input coordinate reference system must only contain a 'Code' and 'Name' keys"
            )

        # get the actual coordinate reference system
        coordinate_reference_system = {
            "Current": value,
            "Previous": self.coordinate_reference_system,
        }

        # update the metadata
        self.metadata = {"Coordinate Reference System": coordinate_reference_system}

    @classmethod
    def create(cls, workspace, **kwargs):
        """
        Function to create an entity.

        :param workspace: Workspace to be added to.
        :param kwargs: List of keyword arguments defining the properties of a class.

        :return entity: Registered Entity to the workspace.
        """
        entity_type_kwargs = (
            {"uid": kwargs.pop("entity_type_uid")}
            if "entity_type_uid" in kwargs
            else {}
        )
        new_object = workspace.create_entity(
            cls,
            entity=kwargs,
            entity_type=entity_type_kwargs,
        )
        return new_object

    @classmethod
    @abstractmethod
    def default_type_uid(cls) -> uuid.UUID | None:
        """Abstract method to return the default type uid for the class."""

    @property
    @abstractmethod
    def entity_type(self):
        """Abstract property to get the entity type of the entity."""

    @classmethod
    def find_or_create_type(cls, workspace: Workspace, **kwargs) -> EntityType:
        """
        Find or create a type instance for a given object class.

        :param workspace: Target :obj:`~geoh5py.workspace.workspace.Workspace`.

        :return: The ObjectType instance for the given object class.
        """
        kwargs["entity_class"] = cls
        return EntityType.find_or_create(workspace, **kwargs)

    @classmethod
    def fix_up_name(cls, name: str) -> str:
        """If the given  name is not a valid one, transforms it to make it valid
        :return: a valid name built from the given name. It simply returns the given name
        if it was already valid.
        """
        # TODO: implement an actual fixup
        #  (possibly it has to be abstract with different implementations per Entity type)
        return name

    @abstractmethod
    def mask_by_extent(
        self, extent: np.ndarray, inverse: bool = False
    ) -> np.ndarray | None:
        """
        Get a mask array from coordinate extent.

        :param extent: Bounding box extent coordinates defined by either:
            - obj:`numpy.ndarray` of shape (2, 3)
            3D coordinate: [[west, south, bottom], [east, north, top]]
            - obj:`numpy.ndarray` of shape (2, 2)
            Horizontal coordinates: [[west, south], [east, north]].
        :param inverse: Return the complement of the mask extent. Default to False

        :return: Array of bool defining the vertices or cell centers
            within the mask extent, or None if no intersection.
        """

    @property
    def metadata(self) -> dict | None:
        """
        Metadata attached to the entity.
        To update the metadata, use the setter method.
        To remove the metadata, set it to None.
        """
        if getattr(self, "_metadata", None) is None:
            value = self.workspace.fetch_metadata(self)
            self._metadata = self.validate_metadata(value)

        return self._metadata

    @metadata.setter
    def metadata(self, value: dict | np.ndarray | bytes | None):
        self._metadata = self.validate_metadata(value)

        if self.on_file:
            self.workspace.update_attribute(self, "metadata")

    def update_metadata(self, value: dict):
        """
        Update the metadata of the entity.

        :param value: Metadata to update.
        """
        metadata = self.metadata if isinstance(self.metadata, dict) else {}

        if isinstance(value, dict):
            value = {**metadata, **value}
            self.metadata = value
        else:
            raise TypeError(
                "Input metadata must be of type dict. "
                f"Provided value of type '{type(value)}'."
            )

    @staticmethod
    def validate_metadata(value) -> dict | None:
        if isinstance(value, np.ndarray):
            value = value[0]

        if isinstance(value, bytes):
            value = str_json_to_dict(value)

        if not isinstance(value, (dict, type(None))):  # remove the metadata
            raise TypeError(
                "Input metadata must be of type dict or None. "
                f"Provided value of type '{type(value)}'."
            )

        return value

    @abstractmethod
    def validate_entity_type(self, entity_type):
        """
        Validate the entity type.
        """

    @property
    def name(self) -> str:
        """
        :obj:`str` Name of the entity
        """
        return self._name

    @name.setter
    def name(self, new_name: str):
        self._name = self.fix_up_name(new_name)

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def on_file(self) -> bool:
        """
        Whether this Entity is already stored on
        :obj:`~geoh5py.workspace.workspace.Workspace.h5file`.
        """
        return self._on_file

    @on_file.setter
    def on_file(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("Attribute 'on_file' must be of type bool.")

        self._on_file = value

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent: EntityContainer):
        current_parent: EntityContainer | None = getattr(self, "_parent", None)

        self._parent = parent
        parent.add_children([self])

        if current_parent is not None and current_parent != self._parent:
            current_parent.remove_children([self])
            self.workspace.save_entity(self)

    @property
    def partially_hidden(self) -> bool:
        """
        Whether this Entity is partially hidden.
        """
        return self._partially_hidden

    @partially_hidden.setter
    def partially_hidden(self, value: bool):
        self._partially_hidden = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def public(self) -> bool:
        """
        Whether this Entity is accessible in the workspace tree and other parts
            of the the user interface in ANALYST.
        """
        return self._public

    @public.setter
    def public(self, value: bool):
        self._public = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def uid(self) -> uuid.UUID:
        return self._uid

    @uid.setter
    def uid(self, uid: str | uuid.UUID):
        uid = str2uuid(uid)

        if not isinstance(uid, uuid.UUID):
            raise TypeError("Input uid must be a string or uuid.UUID.")

        self._uid = uid

    @property
    def visible(self) -> bool:
        """
        Whether the Entity is visible in camera (checked in ANALYST object tree).
        """
        return self._visible

    @visible.setter
    def visible(self, value: bool):
        self._visible = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def workspace(self) -> Workspace:
        """
        :obj:`~geoh5py.workspace.workspace.Workspace` to which the Entity belongs to.
        """
        return self.entity_type.workspace



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid
import weakref
from abc import ABC
from typing import TYPE_CHECKING, Any, TypeVar, cast
from warnings import warn

from ..shared.utils import ensure_uuid


if TYPE_CHECKING:
    from ..workspace import Workspace
    from .entity import Entity

EntityTypeT = TypeVar("EntityTypeT", bound="EntityType")


class EntityType(ABC):
    # pylint: disable=too-many-arguments
    """
    The base class for all entity types.

    :param workspace: The workspace to associate the entity type with.
    :param uid: The unique identifier of the entity type.
    :param description: The description of the entity type.
    :param name: The name of the entity type.
    :param on_file: Return True if the entity is on file.
    """

    _attribute_map = {"Description": "description", "ID": "uid", "Name": "name"}

    def __init__(
        self,
        workspace: Workspace,
        *,
        uid: uuid.UUID | None = None,
        description: str | None = "Entity",
        name: str = "Entity",
        on_file: bool = False,
        **_,
    ):
        self._uid: uuid.UUID = ensure_uuid(uid) if uid is not None else uuid.uuid4()

        self.description = description
        self.name = name
        self.workspace = workspace
        self.on_file = on_file

    @property
    def attribute_map(self) -> dict[str, str]:
        """
        Correspondence map between property names used in geoh5py and
        geoh5.
        """
        return self._attribute_map

    @classmethod
    def convert_kwargs(cls, kwargs: dict[str, Any]) -> dict[str, Any]:
        """
        Convert the kwargs to the geoh5py attribute names.

        :param kwargs: The kwargs to convert.

        :return: The converted kwargs.
        """
        return {
            cls._attribute_map.get(key, key): value for key, value in kwargs.items()
        }

    def copy(self, **kwargs):
        """
        Copy this entity type to another workspace.
        """

        attributes = {
            prop: getattr(self, prop)
            for prop in dir(self)
            if isinstance(getattr(self.__class__, prop, None), property)
            and getattr(self, prop) is not None
            and prop != "attribute_map"
        }

        attributes.update(kwargs)

        if attributes.get("uid") in attributes.get("workspace", self.workspace)._types:  # pylint: disable=protected-access
            del attributes["uid"]

        return self.__class__(**attributes)

    @classmethod
    def create_custom(cls, workspace: Workspace, **kwargs):
        """
        WILL BE  DEPRECATED IN 10.0.0

        Creates a new instance of GroupType for an unlisted custom Group type with a
        new auto-generated UUID.
        """
        warn("This method will be deprecated in 10.0.0. Use the class constructor")
        return cls(workspace, **kwargs)

    @property
    def description(self) -> str | None:
        """
        The description of the entity type.
        """
        return self._description

    @description.setter
    def description(self, description: str | None):
        if not isinstance(description, (str, type(None))):
            raise TypeError(
                f"Description must be a string or None, find {type(description)}"
            )

        self._description = description

        if self.workspace:
            self.workspace.update_attribute(self, "attributes")

    @classmethod
    def find(
        cls: type[EntityTypeT], workspace: Workspace, type_uid: uuid.UUID
    ) -> EntityTypeT | None:
        """
        Finds in the given Workspace the EntityType with the given UUID for
        this specific EntityType implementation class.

        :return: EntityType of None
        """
        return cast(EntityTypeT, workspace.find_type(ensure_uuid(type_uid), cls))

    @classmethod
    def find_or_create(
        cls,
        workspace: Workspace,
        uid: uuid.UUID | None = None,
        entity_class: type[Entity] | None = None,
        **kwargs,
    ):
        """
        Find or creates an EntityType with given uid that matches the given
        Group implementation class.

        It is expected to have a single instance of EntityType in the Workspace
        for each concrete Entity class.

        To find an object, the kwargs must contain an existing 'uid' keyword,
        or a 'entity_class' keyword, containing an object class.

        :param workspace: An active Workspace class
        :param uid: The unique identifier of the entity type.
        :param entity_class: The class of the entity.
        :param kwargs: The attributes of the entity type.

        :return: EntityType
        """
        kwargs = cls.convert_kwargs(kwargs)
        uid = kwargs.pop("uid", uid)

        if entity_class is not None and uid is None:
            uid = entity_class.default_type_uid()

        if uid is not None:
            entity_type = cls.find(workspace, ensure_uuid(uid))
            if entity_type is not None:
                return entity_type

        return cls(workspace, uid=uid, **kwargs)

    @property
    def name(self) -> str:
        """
        The name of the entity type.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        if not isinstance(name, str):
            raise TypeError(f"name must be a string, not {type(name)}")

        self._name = name

        if self.workspace:
            self.workspace.update_attribute(self, "attributes")

    @property
    def on_file(self) -> bool:
        """
        Return True if Entity already present in
        the workspace.
        """
        return self._on_file

    @on_file.setter
    def on_file(self, value: bool):
        if not isinstance(value, bool) and value != 1 and value != 0:
            raise TypeError(f"on_file must be a bool, not {type(value)}")
        self._on_file = bool(value)

    @property
    def uid(self) -> uuid.UUID:
        """
        The unique identifier of an entity, either as stored
        in geoh5 or generated in :func:`~uuid.UUID.uuid4` format.
        """
        return self._uid

    @property
    def workspace(self) -> Workspace:
        """
        The Workspace associated to the object.
        """
        if not hasattr(self, "_workspace"):
            return None  # type: ignore

        _workspace = self._workspace()
        if _workspace is None:
            raise AssertionError("Cannot access the workspace, ensure it is open.")

        return _workspace

    @workspace.setter
    def workspace(self, workspace: Workspace):
        if hasattr(self, "_workspace"):
            raise AssertionError("Cannot change the workspace of an entity type.")
        if not hasattr(workspace, "create_entity"):
            raise TypeError(f"Workspace must be a Workspace, not {type(workspace)}")

        self._workspace = weakref.ref(workspace)
        self.workspace.register(self)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=too-many-ancestors

from __future__ import annotations

import numpy as np
from h5py import special_dtype

from ...data import Data
from ...objects import Drillhole
from .object import ConcatenatedObject
from .property_group import ConcatenatedPropertyGroup


class ConcatenatedDrillhole(ConcatenatedObject, Drillhole):
    def _update_attribute_from_property_group(
        self,
        attributes: dict,
        property_group: ConcatenatedPropertyGroup | str,
    ) -> tuple[dict, ConcatenatedPropertyGroup]:
        """
        Update the input attributes with the depth or from-to values from the property group.

        :param attributes: The attributes to update.
        :param property_group: the property group to use for the update.

        :return: The updated attributes and the property group.
        """

        if isinstance(property_group, str):
            property_group = self.get_property_group(property_group)[0]  # type: ignore
        if not isinstance(property_group, ConcatenatedPropertyGroup):
            raise AttributeError(
                "Input data dictionary must contain a key/value pair of depth data "
                "or contain an 'OBJECT' association. Valid depth keys are 'depth' "
                "and 'from-to'."
            )
        if property_group.property_group_type == "Depth table":
            attributes["depth"] = property_group.depth_.values
        elif property_group.property_group_type == "Interval table":
            attributes["from-to"] = np.c_[
                property_group.from_.values, property_group.to_.values
            ]
        else:
            raise AttributeError(
                "Input data property group must be of type 'Depth table' or 'Interval table'"
            )

        return attributes, property_group

    @property
    def depth_(self) -> list[Data]:
        obj_list = []
        for prop_group in (
            self.property_groups if self.property_groups is not None else []
        ):
            properties = [] if prop_group.properties is None else prop_group.properties
            data = [self.get_data(child)[0] for child in properties]
            if data and "depth" in data[0].name.lower():
                obj_list.append(data[0])

        return obj_list

    @property
    def from_(self) -> list[Data]:
        """
        Depth data corresponding to the tops of the interval values.
        """
        obj_list = []
        for prop_group in (
            self.property_groups if self.property_groups is not None else []
        ):
            properties = [] if prop_group.properties is None else prop_group.properties
            data = [self.get_data(child)[0] for child in properties]
            if len(data) > 0 and "from" in data[0].name.lower():
                obj_list.append(data[0])
        return obj_list

    @property
    def to_(self) -> list[Data]:
        """
        Depth data corresponding to the bottoms of the interval values.
        """
        obj_list = []
        for prop_group in (
            self.property_groups if self.property_groups is not None else []
        ):
            data = [self.get_data(child)[0] for child in prop_group.properties]
            if len(data) > 1 and "to" in data[1].name.lower():
                obj_list.append(data[1])
        return obj_list

    def validate_association(
        self, attributes: dict, property_group=None, collocation_distance=None, **_
    ) -> tuple:
        """
        Validate input drillhole data attributes.

        :param attributes: Dictionary of data attributes.
        :param property_group: Input property group to validate against.
        :param collocation_distance: Threshold on the comparison between existing depth values.
        """
        if collocation_distance is None:
            collocation_distance = attributes.get(
                "collocation_distance", self.default_collocation_distance
            )

        if attributes["name"] in self.get_data_list():
            raise ValueError(
                f"Data with name '{attributes['name']}' already present "
                f"on the drillhole '{self.name}'. "
                "Consider changing the values or renaming."
            )

        if collocation_distance < 0:
            raise UserWarning("Input depth 'collocation_distance' must be >0.")

        if (
            "depth" not in attributes
            and "from-to" not in attributes
            and "association" not in attributes
        ):
            attributes, property_group = self._update_attribute_from_property_group(
                attributes, property_group
            )

        # set a specific nan value if text
        if attributes.get("type") == "TEXT":
            nan_value = ""
        else:
            nan_value = np.nan

        if "depth" in attributes.keys():
            values = attributes.get("values")
            attributes["association"] = "DEPTH"
            property_group = self.validate_depth_data(
                attributes.get("depth"),
                values,
                property_group=property_group,
                collocation_distance=collocation_distance,
            )

            if (
                isinstance(values, np.ndarray)
                and values.shape[0] < property_group.depth_.values.shape[0]
            ):
                attributes["values"] = np.pad(
                    values,
                    (0, property_group.depth_.values.shape[0] - len(values)),
                    constant_values=nan_value,
                )

            del attributes["depth"]

        if "from-to" in attributes.keys():
            values = attributes.get("values")
            attributes["association"] = "DEPTH"
            property_group = self.validate_interval_data(
                attributes.get("from-to"),
                values,
                property_group=property_group,
                collocation_distance=collocation_distance,
            )
            if (
                isinstance(values, np.ndarray)
                and values.shape[0] < property_group.from_.values.shape[0]
            ):
                attributes["values"] = np.pad(
                    values,
                    (0, property_group.from_.values.shape[0] - len(values)),
                    constant_values=nan_value,
                )

            del attributes["from-to"]

        return attributes, property_group

    def validate_depth_data(
        self,
        depth: list | np.ndarray | None,
        values: np.ndarray,
        collocation_distance: float | None = None,
        property_group: str | ConcatenatedPropertyGroup | None = None,
    ) -> ConcatenatedPropertyGroup:
        """
        Compare new and current depth values and reuse the property group if possible.

        :param depth: Sampling depths.
        :param values: Data samples to depths.
        :param collocation_distance: Threshold on the comparison between existing depth values.
        :param property_group: Group for possibly collocated data.

        :return: Augmented property group with name/values added for collocated data
            otherwise newly created property group with name/depth/values added.
        """
        if depth is not None:
            if isinstance(depth, list):
                depth = np.vstack(depth)

            if len(depth) < len(values):
                msg = f"Mismatch between input 'depth' shape{depth.shape} "
                msg += f"and 'values' shape{values.shape}"
                raise ValueError(msg)

        if depth is not None and self.property_groups is not None:
            for group in self.property_groups:
                if (
                    group.depth_ is not None
                    and group.depth_.values.shape[0] == depth.shape[0]
                    and np.allclose(
                        group.depth_.values, depth, atol=collocation_distance
                    )
                ):
                    if isinstance(property_group, str) and group.name != property_group:
                        continue

                    return group

        ind = 0
        label = ""
        if len(self.depth_) > 0:
            ind = len(self.depth_)
            label = f"({ind})"

        if property_group is None:
            property_group = f"depth_{ind}"

        if isinstance(property_group, str):
            out_group: ConcatenatedPropertyGroup = self.fetch_property_group(  # type: ignore
                name=property_group,
                association="DEPTH",
                property_group_type="Depth table",
            )

            if out_group.depth_ is not None:
                if out_group.depth_.values.shape[0] != values.shape[0]:
                    raise ValueError(
                        f"Input values with shape({values.shape[0]}) "
                        f"do not match the from-to intervals of the group '{out_group}' "
                        f"with shape({out_group.depth_.values.shape[0]}). Check values or "
                        f"assign to a new property group."
                    )
                return out_group

        else:
            out_group = property_group

        self.add_data(
            {
                f"DEPTH{label}": {
                    "association": "DEPTH",
                    "values": depth,
                    "primitive_type": "FLOAT",
                    "parent": self,
                    "allow_move": False,
                    "allow_delete": False,
                },
            },
            out_group,
        )

        return out_group

    def validate_interval_data(
        self,
        from_to: list | np.ndarray | None,
        values: np.ndarray,
        collocation_distance=1e-4,
        property_group: str | ConcatenatedPropertyGroup | None = None,
    ) -> ConcatenatedPropertyGroup:
        """
        Compare new and current depth values and reuse the property group if possible.
        Otherwise a new property group is added.

        :param from_to: Array of from-to values.
        :param values: Data values to be added on the from-to intervals.
        :param collocation_distance: Threshold on the comparison between existing depth values.
        :param property_group: Property group name.

        :return A ConcatenatedPropertyGroup with the matched values.
        """

        if from_to is not None:
            if isinstance(from_to, list):
                from_to = np.vstack(from_to)
                if from_to.shape[0] == 2:
                    from_to = from_to.T

            assert from_to.shape[0] >= len(values), (
                f"Mismatch between input 'from_to' shape{from_to.shape} "
                + f"and 'values' shape{values.shape}"
            )
            assert from_to.shape[1] == 2, "The `from-to` values must have shape(*, 2)"

        if (
            from_to is not None
            and property_group is None
            and self.property_groups is not None
        ):
            for p_g in self.property_groups:
                if (
                    isinstance(p_g.from_, Data)
                    and p_g.from_.values.shape[0] == from_to.shape[0]
                    and np.allclose(
                        np.c_[p_g.from_.values, p_g.to_.values],
                        from_to,
                        atol=collocation_distance,
                    )
                ):
                    return p_g

        ind = 0
        label = ""
        if len(self.from_) > 0:
            ind = len(
                list(set(self.from_))
            )  # todo: from_ return the same value x time why?
            label = f"({ind})"

        if property_group is None:
            property_group = f"Interval_{ind}"

        if isinstance(property_group, str):
            out_group: ConcatenatedPropertyGroup = self.fetch_property_group(  # type: ignore
                name=property_group,
                association="DEPTH",
                property_group_type="Interval table",
            )
        else:
            out_group = property_group

        if out_group.from_ is not None:
            if out_group.from_.values.shape[0] != values.shape[0]:
                raise ValueError(
                    f"Input values with shape({values.shape[0]}) "
                    f"do not match the from-to intervals of the group '{out_group}' "
                    f"with shape({out_group.from_.values.shape[0]}). Check values or "
                    f"assign to a new property group."
                )
            return out_group

        self.add_data(
            {
                f"FROM{label}": {
                    "association": "DEPTH",
                    "values": from_to[:, 0],
                    "primitive_type": "FLOAT",
                    "parent": self,
                    "allow_move": False,
                    "allow_delete": False,
                },
                f"TO{label}": {
                    "association": "DEPTH",
                    "values": from_to[:, 1],
                    "primitive_type": "FLOAT",
                    "parent": self,
                    "allow_move": False,
                    "allow_delete": False,
                },
            },
            out_group,
        )

        return out_group

    def post_processing(self):
        """Bypass sort_depths from previous version."""

    def format_survey_values(self, values: list | np.ndarray) -> np.recarray:
        """
        Reformat the survey values as structured array with the right shape.
        """
        if isinstance(values, (list, tuple)):
            values = np.array(values, ndmin=2)

        if isinstance(values, np.ndarray):
            values = values.T.tolist()

        dtype = [("Depth", "<f4"), ("Azimuth", "<f4"), ("Dip", "<f4")]

        if (
            "Surveys" in self.concatenator.data
            and len(self.concatenator.data["Surveys"].dtype) == 4
        ):
            dtype = self.concatenator.data["Surveys"].dtype

        if len(values) not in [3, 4]:
            raise ValueError("'surveys' requires an ndarray of shape (*, 3) or (*, 4)")

        if len(values) == 3 and len(dtype) == 4:
            values += [np.array([b""] * len(values[0]), dtype=special_dtype(vlen=str))]
        elif len(values) == 4 and len(dtype) == 3:
            values = values[:-1]

        array_values = np.core.records.fromarrays(values, dtype=dtype)

        return array_values



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from abc import ABC, abstractmethod
from numbers import Real

import numpy as np

from .object_base import ObjectBase


ORIGIN_TYPE = np.dtype([("x", float), ("y", float), ("z", float)])


class GridObject(ObjectBase, ABC):
    """
    Base class for object with centroids.

    :param origin: Origin of the object.
    :param rotation: Rotation angle (clockwise) about the vertical axis.
    """

    _attribute_map = ObjectBase._attribute_map.copy()

    def __init__(
        self,
        origin: np.ndarray | tuple = (0.0, 0.0, 0.0),
        rotation: float = 0.0,
        **kwargs,
    ):
        self._centroids: np.ndarray | None = None

        super().__init__(**kwargs)

        self.origin = origin
        self.rotation = rotation

    @property
    @abstractmethod
    def centroids(self) -> np.ndarray:
        """
        Cell center locations in world coordinates of shape (n_cells, 3).
        """

    @property
    def n_cells(self) -> int:
        """
        Total number of cells
        """
        return int(np.prod(self.shape))

    @property
    def rotation(self) -> float:
        """
        Clockwise rotation angle (degree) about the vertical axis.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, value: np.ndarray | Real):
        if isinstance(value, Real):
            value = np.r_[value]

        if not isinstance(value, np.ndarray) or value.shape != (1,):
            raise TypeError("Rotation angle must be a float of shape (1,)")

        self._centroids = None
        self._rotation = value.astype(float).item()

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def origin(self) -> np.ndarray:
        """
        Coordinates of the origin, shape (3, ).
        """
        return self._origin

    @origin.setter
    def origin(self, values: np.ndarray | list | tuple):
        if isinstance(values, (list, tuple)):
            values = np.array(values)

        if not isinstance(values, (np.ndarray, np.void)):
            raise TypeError(
                "Attribute 'origin' must be a list, tuple or numpy array. "
                f"Object of type {type(values)} provided."
            )

        if np.issubdtype(values.dtype, np.number):
            if len(values) != 3:
                raise ValueError(
                    "Attribute 'origin' must be a list or array of shape (3,). "
                    f"Array of shape {values.shape} provided."
                )

            values = np.asarray(tuple(values), dtype=ORIGIN_TYPE)

        if values.dtype != np.dtype(ORIGIN_TYPE):
            raise ValueError(f"Array of 'origin' must be of dtype = {ORIGIN_TYPE}")

        self._centroids = None

        if getattr(self, "_origin", None) is not None and self.on_file:
            self.workspace.update_attribute(self, "attributes")

        self._origin = values

    @property
    @abstractmethod
    def shape(self) -> np.ndarray:
        """
        Cell center locations in world coordinates.
        """



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import cast

import numpy as np

from ...data import NumericData
from ...objects import DrapeModel
from ...workspace import Workspace
from .base import BaseMerger


class DrapeModelMerger(BaseMerger):
    _type: type = DrapeModel

    @classmethod
    def create_object(
        cls, workspace: Workspace, input_entities: list, **kwargs
    ) -> DrapeModel:
        """
        Create a new DrapeModel from a list of input DrapeModels.

        It merges the cells together and create a new object with the merged cells.

        :param workspace: The workspace to create the object in.
        :param input_entities: The list of input entities to merge together.
        :param kwargs: The kwargs to pass to the object creation.

        :return: The newly created object merged from input_entities.
        """

        layers: list = []
        prisms: list = []
        previous_prism: int = 0
        previous_layer: int = 0
        ghost_prism: np.ndarray = np.array([])
        ghost_layer: np.ndarray = np.array([])

        for input_entity in input_entities:
            # get the values of the entity
            temp_prisms: np.ndarray = input_entity.prisms
            temp_layers: np.ndarray = input_entity.layers

            if len(temp_prisms) < 2:
                raise ValueError("All DrapeModel entities must have at least 2 prisms.")

            # get the first ghost
            if ghost_prism.size > 0:
                # append last ghost prism and layer
                prisms.append(ghost_prism)
                layers.append(ghost_layer)

                # create the first ghost point
                ghost_prism, ghost_layer = cls._ghost_point(
                    temp_prisms[0],
                    temp_prisms[1],
                    previous_prism - 1,
                    previous_layer - 1,
                )

                # append first ghost prism and layer
                prisms.append(ghost_prism)
                layers.append(ghost_layer)

            # add the entity prisms and layers to the list
            temp_prisms[:, -2] += previous_prism
            temp_layers[:, 0] += previous_layer

            previous_prism = temp_prisms[-1, -1] + temp_prisms[-1, -2] + 2
            previous_layer = temp_layers[-1, 0] + 3

            prisms.append(temp_prisms)
            layers.append(temp_layers)

            # create a last ghost prism and layer
            ghost_prism, ghost_layer = cls._ghost_point(
                temp_prisms[-1], temp_prisms[-2], previous_prism - 2, previous_layer - 2
            )

        # create an object
        output = cls._type.create(  # type: ignore
            workspace, prisms=np.vstack(prisms), layers=np.vstack(layers), **kwargs
        )

        return output

    @staticmethod
    def _ghost_point(
        point: np.ndarray,
        mirror: np.ndarray,
        previous_prism: int,
        previous_layer: int,
    ) -> np.ndarray | np.ndarray:
        """
        Create a ghost point prism and layer based on two points.
        :param point: The point to create the ghost from.
        :param mirror: The mirror point to create the ghost from.
        :param previous_prism: The ID of the previous prism.
        :param previous_layer: The ID of the previous layer.
        :return: A prism and a layer for the ghost point.
        """

        # create a mirrored point for last layer
        ghost_prism: np.ndarray = point.copy()
        ghost_layer: np.ndarray = np.empty(3)
        ghost_prism[:3] = 2 * point[:3] - mirror[:3]

        ghost_prism[3] = previous_prism
        ghost_prism[4] = 1

        ghost_layer[0] = previous_layer
        ghost_layer[1] = 0
        ghost_layer[2] = ghost_prism[2]

        return np.expand_dims(ghost_prism, 0), np.expand_dims(ghost_layer, 0)

    @classmethod
    def merge_data(
        cls,
        out_entity,
        input_entities: list[DrapeModel],
    ):
        super().merge_data(out_entity, input_entities)

        ind_map: list = []
        data_count: int = 0
        n_values: int = out_entity.n_cells - (len(input_entities) - 1) * 2
        for ghost, input_entity in enumerate(input_entities):
            if input_entity.n_cells is None:
                continue
            ind_map += [
                np.arange(data_count, data_count + input_entity.n_cells),
                [n_values + ghost * 2, n_values + ghost * 2 + 1],
            ]
            data_count += input_entity.n_cells

        # get all the values in the output entity
        for data in out_entity.children:
            if not isinstance(data, NumericData) or data.values is None:
                continue
            data.values = data.values[np.hstack(ind_map[:-1])]

    @classmethod
    def validate_structure(cls, input_entity: DrapeModel):
        """
        Validate the input entity structure and raises error if incompatible.
        :param input_entity: the input entity to validate.
        """
        # verify if the input entity have prism and layers
        if not (
            isinstance(cast(DrapeModel, input_entity).prisms, np.ndarray)
            and isinstance(cast(DrapeModel, input_entity).layers, np.ndarray)
        ):
            raise AttributeError("All entities must have prisms and layers.")



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from warnings import warn

import numpy as np

from .data import Data, PrimitiveTypeEnum
from .data_association_enum import DataAssociationEnum


class NumericData(Data, ABC):
    """
    Data container for floats values
    """

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.INVALID

    @property
    @abstractmethod
    def ndv(self):
        """No-data-value"""

    @property
    def ndv_values(self) -> np.ndarray | None:
        """
        Data with nan replaced by ndv
        """
        if self.values is None:
            return None
        values = deepcopy(self.values)
        values[np.isnan(values)] = self.ndv

        return values

    @abstractmethod
    def format_type(self, values: np.ndarray) -> np.ndarray:
        """
        Check if the type of values is valid and convert it to right dtype.
        :param values: numpy array to modify.
        :return: the formatted values.
        """

    def format_length(self, values: np.ndarray) -> np.ndarray:
        """
        Check for possible mismatch between the length of values
        :param values: the values to check.
        :return: the values with the right length.
        """

        if self.n_values is None:
            return values

        if len(values) < self.n_values:
            full_vector = np.ones(self.n_values, dtype=values.dtype) * self.nan_value
            full_vector[: len(np.ravel(values))] = np.ravel(values)
            return full_vector

        if (
            len(values) > self.n_values
            and self.association is not DataAssociationEnum.OBJECT
        ):
            raise ValueError(
                f"Input 'values' of shape({self.n_values},) expected. "
                f"Array of shape{values.shape} provided.)"
            )

        return values

    def validate_values(self, values: np.ndarray | None) -> np.ndarray:
        """
        Check for possible mismatch between the length of values
        stored and the expected number of cells or vertices.

        :param values: Array of values to check

        :returns: values: An array of float values of length n_values or None
        """
        if values is None:
            return values

        if not isinstance(values, np.ndarray):
            raise TypeError("Input 'values' must be a numpy array.")

        if values.ndim > 1:
            values = np.ravel(values)
            warn("Input 'values' converted to a 1D array.")

        values = values.astype(float)

        # change nan values to nan_value
        values[np.isnan(values)] = self.nan_value

        # check the length of the values
        values = self.format_length(values)

        # check the value type
        values = self.format_type(values)

        return values



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=unused-import, cyclic-import
# flake8: noqa

from .block_model import BlockModel
from .cell_object import CellObject
from .curve import Curve
from .drape_model import DrapeModel
from .drillhole import Drillhole
from .geo_image import GeoImage
from .grid2d import Grid2D
from .integrator import IntegratorPoints, NeighbourhoodSurface
from .label import Label
from .notype_object import NoTypeObject
from .object_base import ObjectBase
from .object_type import ObjectType
from .octree import Octree
from .points import Points
from .slicer import Slicer
from .surface import Surface
from .surveys.direct_current import CurrentElectrode, PotentialElectrode
from .surveys.electromagnetics.airborne_fem import (
    AirborneFEMReceivers,
    AirborneFEMTransmitters,
)
from .surveys.electromagnetics.airborne_tem import (
    AirborneTEMReceivers,
    AirborneTEMTransmitters,
)
from .surveys.electromagnetics.base import (
    AirborneEMSurvey,
    FEMSurvey,
    LargeLoopGroundEMSurvey,
    MovingLoopGroundEMSurvey,
    TEMSurvey,
)
from .surveys.electromagnetics.ground_fem import (
    LargeLoopGroundFEMReceivers,
    LargeLoopGroundFEMTransmitters,
    MovingLoopGroundFEMReceivers,
    MovingLoopGroundFEMTransmitters,
)
from .surveys.electromagnetics.ground_tem import (
    LargeLoopGroundTEMReceivers,
    LargeLoopGroundTEMTransmitters,
    MovingLoopGroundTEMReceivers,
    MovingLoopGroundTEMTransmitters,
)
from .surveys.electromagnetics.magnetotellurics import MTReceivers
from .surveys.electromagnetics.tipper import TipperBaseStations, TipperReceivers
from .surveys.magnetics import AirborneMagnetics



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from uuid import UUID

from .data import Data
from .primitive_type_enum import PrimitiveTypeEnum


class GeometricDataConstants(Data):
    """
    Base class for geometric data constants.

    :param allow_move: Defaults coordinate to remain on object.
    :param visible: Defaults to not visible.
    """

    _TYPE_UID: UUID

    def __init__(
        self,
        allow_move=False,
        visible=False,
        allow_delete=False,
        allow_rename=False,
        **kwargs,
    ):
        super().__init__(
            allow_move=allow_move,
            allow_delete=allow_delete,
            allow_rename=allow_rename,
            visible=visible,
            **kwargs,
        )

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.GEOMETRIC

    def validate_values(self, values: None) -> None:
        """
        Validate values for GeometricDataConstants.
        """
        if values is not None:
            raise TypeError(
                f"GeometricDataConstants does not accept values. Got {values}."
            )

        return values
